import pandas as pd
import texttable as tt
from fastprogress import progress_bar as pb

from kts.feature_selection.selector import BuiltinImportance
from kts.util.misc import hash_str
from kts.eda.importance import plot_importances
from kts.modelling.mixins import ArithmeticMixin


class Experiment(ArithmeticMixin):
    """ """
    def __init__(
            self,
            pipeline,
            oof,
            score,
            std,
            description,
            validator,
            feature_list,
            helper_list,
    ):
        self.pipeline = pipeline
        self.model = self.pipeline.models[0].model.model  # TODO: test out
        self.model_name = self.model.__class__.__name__
        self.parameters = self.model.get_params()
        self.models = [model.model.model for model in self.pipeline.models]
        self.featureset = self.pipeline.models[0].model.featureslice.featureset
        self.tie_featuresets()
        self.oof = oof
        self.score = score
        self.std = std
        self.identifier = hash_str(f"{pipeline.__name__}")[:6].upper()
        self.oof.columns = [
            col_name.replace("prediction", self.identifier)
            for col_name in self.oof.columns
        ]
        self.__doc__ = description if description is not None else "no description"
        self.__name__ = f"{self.identifier}-{round(score, 4)}-{pipeline.__name__}"
        self.validator = validator
        self.features = list(feature_list)
        self.helpers = list(helper_list)

    def __str__(self):
        string = f"({round(self.score, 5)}, std:{round(self.std, 3)}: \n\tModel: {self.pipeline.__name__})"
        return string

    def predict(self, df):
        """

        Args:
          df: 

        Returns:

        """
        return self.pipeline.predict(df)

    def __repr__(self):
        fields = {
            "Score":
            f"{round(self.score, 7)}, std: {round(self.std, 7)} ({self.validator.metric.__name__})",
            "Identifier": self.identifier,
            "Description": self.__doc__,
            # 'Model': self.model_name + f'\tx{len(self.models)}',
            # 'Model parameters': self.parameters,
            "Model": f"{self.model.__name__}\t x{len(self.models)}",
            "|- source ": self.model.
            source,  # be careful with refactoring: if you remove this space,
            "FeatureSet": self.featureset.__name__,  #
            "|- description": self.featureset.__doc__,  #
            "|- source": self.featureset.__repr__(
            ),  # both "source" rows will be considered identical
            "Splitter": self.validator.splitter,
        }

        table = tt.Texttable(max_width=80)
        for field in fields:
            table.add_row([field, fields[field]])
        return table.draw()

    def as_dict(self):
        """ """
        fields = {
            "Score": self.score,
            "std": self.std,
            "ID": self.identifier,
            "Model": self.model.__name__,
            "FS": self.featureset.__name__,
            "Description": self.__doc__,
            "FS description": self.featureset.__doc__,
            "Model source": self.model.source,
            "FS source": self.featureset.__repr__(),
            "Splitter": repr(self.validator.splitter),
        }
        for field in fields:
            fields[field] = [fields[field]]
        return fields

    def as_df(self):
        """ """
        return pd.DataFrame(self.as_dict()).set_index("ID")

    def feature_importances(self,
                            plot=False,
                            importance_calculator=BuiltinImportance(),
                            **kw):
        """

        Args:
          plot:  (Default value = False)
          importance_calculator:  (Default value = BuiltinImportance())
          **kw: 

        Returns:

        """
        if plot:
            return plot_importances(self,
                                    calculator=importance_calculator,
                                    **kw)
        res = pd.DataFrame()
        for w_model in pb(self.pipeline.models):
            fsl = w_model.model.featureslice
            model = w_model.model.model
            importances = importance_calculator.calc(model, fsl, self)
            importances = {k: [v] for k, v in importances.items()}
            res = res.append(pd.DataFrame(importances), ignore_index=True)
        return res

    def set_df(self, df_input):
        """

        Args:
          df_input: 

        Returns:

        """
        self.tie_featuresets()
        self.featureset.set_df(df_input)

    def tie_featuresets(self):
        """ """
        for i in range(len(self.pipeline.models)):
            self.pipeline.models[i].model.featureslice.featureset = self.featureset



from kts.api.decorators import preview, register, deregister, dropper, selector, helper
from kts.api.feature import FeatureSet
from kts.api.feature import feature_list as features
from kts.api.helper import helper_list as helpers
from kts.core.backend.memory import cache
from kts.core.backend.memory import save, load, ls, remove, rm
from kts.core.dataframe import DataFrame as KTDF
from kts.core.dataframe import link
from kts.modelling.custom_model import CustomModel
from kts.modelling.stacking import stack
from kts.validation.leaderboard import leaderboard
from kts.validation.validator import Validator

lb = leaderboard



from copy import deepcopy

import numpy as np
import pandas as pd
from fastprogress import master_bar
from fastprogress.fastprogress import IN_NOTEBOOK

from kts.api.feature import FeatureSet
from kts.modelling.pipeline import Pipeline
from kts.util.misc import SourceMetaClass
from kts.validation.experiment import Experiment
from kts.validation.leaderboard import leaderboard
from kts.modelling.mixins import Ensemble, Model
from kts.api.feature import feature_list
from kts.api.helper import helper_list


if IN_NOTEBOOK:
    from IPython.display import HTML


class Validator(metaclass=SourceMetaClass):
    """ """
    def __init__(self, splitter, metric):
        self.splitter = splitter
        self.metric = metric

    def split(self, splitter, featureset):
        """

        Args:
          splitter: 
          featureset: 

        Returns:

        """
        y = featureset.target.values
        for idx_train, idx_test in splitter.split(y, y):
            yield idx_train, idx_test

    def update_oof(self, oof, weights, y_pred, idx_test, full_target):
        """

        Args:
          oof: 
          weights: 
          y_pred: 
          idx_test: 
          full_target: 

        Returns:

        """
        if oof is None or weights is None:
            oof = weights = np.zeros(shape=(full_target.shape[0], ) +
                                     y_pred.shape[1:],
                                     dtype=np.float)
        oof[idx_test] = (weights[idx_test] * oof[idx_test] +
                         y_pred) / (weights[idx_test] + 1)
        weights[idx_test] += 1
        return oof, weights

    def oof_to_df(self, oof, featureset):
        """

        Args:
          oof: 
          featureset: 

        Returns:

        """
        if len(oof.shape) == 1 or oof.shape[1] == 1:
            res = pd.DataFrame({"prediction": oof})
        else:
            res = pd.DataFrame(
                {f"prediction_{i}": oof[:, i]
                 for i in range(oof.shape[1])})
        res.set_index(featureset.target.index, inplace=True)
        return res

    def evaluate(self, y_true, y_pred):
        """

        Args:
          y_true: 
          y_pred: 

        Returns:

        """
        return self.metric(y_true, y_pred)

    def get_ensemble_name(self, model, featureset):
        """

        Args:
          model: 
          featureset: 

        Returns:

        """
        return f"{model.__name__}_x{self.splitter.get_n_splits()}-{featureset.__name__}"

    def __repr__(self):
        return f"Validator({self.splitter}, {self.metric.__name__})"

    def score(
            self,
            model: Model,
            featureset: FeatureSet,
            description: str = None,
            desc: str = None,
            **fit_params,
    ):
        """

        Args:
          model: 
          featureset:
          description:  (Default value = None)
          desc:  (Default value = None)
          **fit_params: 

        Returns:

        """
        if desc is not None and description is not None:
            raise ValueError(
                "desc is an alias of description. You can't use both")
        if desc is not None:
            description = desc
        pipelines = []
        scores = []
        oof = weights = None
        ensemble_name = self.get_ensemble_name(model, featureset)
        mb = master_bar(
            self.split(self.splitter, featureset),
            total=self.splitter.get_n_splits(),
            total_time=IN_NOTEBOOK,
        )
        mb.write(f"Validation of {ensemble_name}:")
        for idx_train, idx_test in mb:
            c_model = deepcopy(model)
            fsl = featureset.slice(idx_train, idx_test)
            pl = Pipeline(c_model, fsl)
            pl.fit(masterbar=mb, **fit_params)
            y_pred = pl.predict(idx_test, masterbar=mb)
            y_true = featureset.target.values[idx_test]
            pipelines.append(pl)
            scores.append(self.evaluate(y_true, y_pred))
            oof, weights = self.update_oof(oof, weights, y_pred, idx_test,
                                           featureset.target.values)
        final_ensemble = Ensemble(pipelines)
        final_ensemble = final_ensemble / len(pipelines)
        final_ensemble.__name__ = ensemble_name
        score = np.mean(scores)
        std = np.std(scores)
        oof = self.oof_to_df(oof, featureset)

        exp = Experiment(
            pipeline=final_ensemble,
            oof=oof,
            score=score,
            std=std,
            description=description,
            validator=self,
            feature_list=list(feature_list),
            helper_list=list(helper_list),
        )
        if IN_NOTEBOOK:
            mb.text = f"ID: {exp.identifier}<p> Score: {score}<p>" + mb.text
            mb.out.update(HTML(mb.text))
        leaderboard.register(exp)
        return {"score": score, "id": exp.identifier}



from kts.modelling.mixins import Model
from kts.util.misc import SourceMetaClass


class CustomModelSourceMetaClass(SourceMetaClass):
    """ """
    def check_methods(methods):
        """

        Args:
          methods:

        Returns:

        """
        required_methods = ["get_short_name", "get_tracked_params"]
        for meth in required_methods:
            assert (meth in methods), f"Method .{meth}() is required to define a custom model"


class CustomModel(Model, metaclass=CustomModelSourceMetaClass):
    """ """
    def get_short_name(self):
        """ """
        return "custom_model"

    def get_tracked_params(self):
        """ """
        return []

    def preprocess(self, X, y):
        """Preprocess input before feeding it into model

        Args:
          X: np.array
          y: np.array or None (fitting or inference)

        Returns:
          X_processed, y_processed)

        """
        return X, y


import numpy as np
import pandas as pd

from kts.core.base_constructors import wrap_stl_function, empty_like
from kts.util.misc import is_helper
from kts.zoo.cluster import KMeansFeaturizer


def make_ohe(cols, sep="_ohe_"):
    """

    Args:
      cols:
      sep:  (Default value = '_ohe_')

    Returns:

    """
    def __make_ohe(df):
        return pd.get_dummies(df[cols], prefix_sep=sep, columns=cols)

    return wrap_stl_function(make_ohe, __make_ohe)


def target_encoding(cols, target_cols, aggregation="mean", sep="_te_"):
    """Template for creating target encoding FeatureConstructors

    Args:
      cols: columns to encode
      target_cols: columns which values are aggregated
      aggregation: name of built-in pandas aggregation or @helper-function (Default value = 'mean')
      sep: string separator used for column naming (Default value = '_te_')

    Returns:
      FeatureConstructor: df -> features

    """
    if type(cols) != list:
        cols = [cols]
    if type(target_cols) != list:
        target_cols = [target_cols]
    if isinstance(aggregation, str):
        aggregation_name = aggregation
    elif is_helper(aggregation):
        aggregation_name = aggregation.__name__
    else:
        raise TypeError(
            "aggregation parameter should be either string or helper")

    def __target_encoding(df):
        res = empty_like(df)
        for target_col in target_cols:
            for col in cols:
                if df.train:
                    enc = df.groupby(col)[target_col].agg(aggregation)
                    global_answer = df[target_col].agg(aggregation)
                    df.encoders[
                        f"__te_{col}_{target_col}_{aggregation_name}"] = enc
                    df.encoders[
                        f"__te_{col}_{target_col}_{aggregation_name}_global_answer"] = global_answer
                else:
                    enc = df.encoders[
                        f"__te_{col}_{target_col}_{aggregation_name}"]
                    global_answer = df.encoders[
                        f"__te_{col}_{target_col}_{aggregation_name}_global_answer"]
                res[f"{col}{sep}{target_col}_{aggregation_name}"] = (
                    df[col].map(enc).astype(np.float))
                res[f"{col}{sep}{target_col}_{aggregation_name}"].fillna(
                    global_answer, inplace=True)
        return res

    return wrap_stl_function(target_encoding, __target_encoding)


def target_encode_list(cols, target_col, aggregation="mean", prefix="me_list_"):
    """

    Args:
      cols:
      target_col:
      aggregation:  (Default value = 'mean')
      prefix:  (Default value = 'me_list_')

    Returns:

    """
    def __target_encode_list(df):
        res = empty_like(df)
        filler = [np.nan]
        for col in cols:
            if df.train:
                enc = ((pd.DataFrame(
                    [i if type(i) == list else [i] for i in df[col].tolist()],
                    index=df[target_col],
                ).stack().reset_index(name=col)[[
                    col, target_col
                ]]).groupby(col).agg(aggregation))
                df.encoders["__me_list_" + col] = enc
            else:
                enc = df.encoders["__me_list_" + col]
            res[prefix + col] = df[col].apply(lambda x: enc.loc[filter(
                lambda i: i in enc.index, x)].values.flatten()
                                              if type(x) == list else filler)
        return res

    return wrap_stl_function(target_encode_list, __target_encode_list)


def kmeans_encoding(cols,
                    n_clusters,
                    target_col=None,
                    target_importance=5.0,
                    prefix="km_",
                    **kwargs):
    """

    Args:
      cols:
      n_clusters:
      target_col:  (Default value = None)
      target_importance:  (Default value = 5.0)
      prefix:  (Default value = 'km_')
      **kwargs:

    Returns:

    """
    def __kmeans_encoding(df):
        res = empty_like(df)
        res_column_name = (
            f"{prefix}{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}"
        )
        if df.train:
            encoder = KMeansFeaturizer(k=n_clusters,
                                       target_scale=target_importance,
                                       **kwargs)
            if target_col:
                res[res_column_name] = encoder.fit_transform(
                    df[cols].values, df[target_col].values)
            else:
                res[res_column_name] = encoder.fit_transform(df[cols].values)
            df.encoders[
                f"__kmeans_feat_{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}"] = encoder
        else:
            encoder = df.encoders[
                f"__kmeans_feat_{n_clusters}_{list_hash(cols, 5)}_{round(target_importance, 4)}"]
            res[res_column_name] = encoder.transform(df[cols].values)
        return res

    return wrap_stl_function(kmeans_encoding, __kmeans_encoding)


import inspect

from kts import config


def get_source(function):
    """Gets source code of a function. Ignores decorators starting with @.

    Args:
      function: function

    Returns:
      source code

    """
    src = inspect.getsource(function)
    if src[0] == "@":
        src = src[src.find("\n") + 1:]
    return src


def shorten(source):
    """

    Args:
      source: 

    Returns:

    """
    if len(source) <= config.MAX_LEN_SOURCE:
        return source
    # source = source.replace('columns=', '')
    left = source[:config.MAX_LEN_SOURCE // 2]
    left = left[:left.rfind(" ")]
    right = source[-config.MAX_LEN_SOURCE // 2:]
    right = right[right.find(" "):]
    return left + " ..." + right




from warnings import warn

import kts.stl.misc
from kts.validation.leaderboard import leaderboard as lb
from kts.validation.split import Refiner
from kts.validation.validator import Validator


def assert_splitters(exps):
    """

    Args:
      exps: 

    Returns:

    """
    all_splitters = set()
    for exp in exps:
        all_splitters.add(repr(exp.splitter))
        if len(all_splitters) > 1:
            raise Exception(
                f"Experiment {repr(exp.identifier)} has {exp.splitter} instead of {all_splitters.pop()}"
            )


def assert_metrics(exps):
    """

    Args:
      exps: 

    Returns:

    """
    all_metrics = set()
    for exp in exps:
        if "source" in dir(exp.validator.metric):
            all_metrics.add(exp.validator.metric.source)
        else:
            all_metrics.add(exp.validator.metric.__name__)
    if len(all_metrics) > 1:
        warn(
            f"Different metrics were used for scoring provided experiments: {all_metrics}."
            f" The first one will be used unless you specify it explicitly.")


def stack(ids,
          safe=True,
          inner_splitter=None,
          metric=None,
          validator_class=Validator):
    """

    Args:
      ids: 
      safe:  (Default value = True)
      inner_splitter:  (Default value = None)
      metric:  (Default value = None)
      validator_class:  (Default value = Validator)

    Returns:

    """
    experiments = lb[ids]
    if safe:
        assert_splitters(experiments)
    outer_splitter = experiments[0].validator.splitter
    assert_metrics(experiments)
    if inner_splitter is None:
        inner_splitter = experiments[0].validator.splitter
    refiner = Refiner(outer_splitter, inner_splitter)
    if metric is None:
        metric = experiments[0].validator.metric
    fc_stack = kts.stl.misc.stack(ids)
    val_stack = validator_class(refiner, metric)
    return val_stack, fc_stack



import os
import shutil


def find_root_dir():
    i = 0
    while os.path.realpath("../" * i) != "/":
        if os.path.exists("../" * i + "kts_config.py"):
            if i != 0:
                return "../" * i
            else:
                return "./"
        i += 1
    return False


def get_mode():
    if find_root_dir():
        # config.STORAGE_PATH = parse(find_root_dir() + '.kts')['STORAGE_PATH']
        cache_mode = "disk_and_ram"
        cache_policy = "everything"
        root_dir = find_root_dir()
    else:
        # warn("Couldn't find existing kts project. Setting kaggle-mode")
        cache_mode = "ram"
        cache_policy = "service"
        root_dir = "."
    return cache_mode, cache_policy, root_dir


def list_files(startpath):
    for root, dirs, files in os.walk(startpath):
        level = root.replace(startpath, '').count(os.sep)
        indent = ' ' * 4 * level
        print('{}{}/'.format(indent, os.path.basename(root)))
        subindent = ' ' * 4 * (level + 1)
        for f in files:
            print('{}{}'.format(subindent, f))


def check_existence(paths):
    """
    Checks necessity of clearing the folder.
    :param paths: list of directories
    :return: True if at least one directory exists, False otherwise
    """
    for path in paths:
        if os.path.isdir(path):
            return True
    return False


def clear_all():
    """
    Clears current folder.
    :return:
    """
    folder = './'
    for the_file in os.listdir(folder):
        file_path = os.path.join(folder, the_file)
        try:
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print(e)


DEFAULT_CONFIG = f"""\
# Cache mode defines resources used for caching:
# - disk         -- use only disk space, no RAM is used. Choose it if you don't have much RAM.
# - ram          -- use only RAM space. Best for kaggle kernels.
# - disk_and_ram -- use both. The fastest option. Best for local usage. Default. 
cache_mode = 'disk_and_ram'  # "disk", "disk_and_ram", "ram"

# Cache policy defines which types of files will be saved.
# - everything   -- cache everything including feature constructor calls. Default.
# - service      -- only service files are saved. No feature computation speedup. 
#                   Use if you're not lucky with your resources.
cache_policy = 'everything'  # "everything", "service"

# Full path of storage.
# DO NOT ERASE
storage_path = '{os.getcwd()}/storage/'

# Task goal: whether greater is better or not
GOAL = 'MAXIMIZE' # or 'MINIMIZE'
"""


def create_config():
    with open('kts_config.py', 'w') as f:
        f.write(DEFAULT_CONFIG)



from collections import MutableSequence

from kts.core.backend import memory


class HelperList(MutableSequence):
    """ """
    def __init__(self):
        self.full_name = "kts.api.helper.helper_list"  # such a hardcode
        self.names = [self.full_name]
        while self.names[-1].count("."):
            self.names.append(self.names[-1][self.names[-1].find(".") + 1:])
        self.names.append("kts.helpers")
        while self.names[-1].count("."):
            self.names.append(self.names[-1][self.names[-1].find(".") + 1:])
        self.objects = []
        self.name_to_idx = dict()

    def recalc(self):
        """ """
        self.objects = []
        self.name_to_idx = dict()
        names = [
            obj for obj in memory.cache.cached_objs()
            if obj.endswith("_helper")
        ]
        for idx, name in enumerate(names):
            functor = memory.cache.load_obj(name)
            self.objects.append(functor)
            self.name_to_idx[functor.__name__] = idx

    def __repr__(self):
        self.recalc()
        string = f"[{', '.join([f.__name__ for f in self.objects])}]"
        return string

    def __getitem__(self, key):
        self.recalc()
        if type(key) in [int, slice]:
            return self.objects[key]
        elif type(key) == str:
            return self.objects[self.name_to_idx[key]]
        else:
            raise TypeError("Index should be int, slice or str")

    def __delitem__(self, key):
        raise AttributeError("This object is read-only")

    def __setitem__(self, key, value):
        raise AttributeError("This object is read-only")

    def insert(self, key, value):
        """

        Args:
          key: 
          value: 

        Returns:

        """
        raise AttributeError("This object is read-only")

    def define_in_scope(self, global_scope):
        """

        Args:
          global_scope: 

        Returns:

        """
        self.recalc()
        for func in self.name_to_idx:
            for name in self.names:
                try:
                    exec(f"{func} = {name}['{func}']", global_scope)
                    break
                except BaseException:
                    pass

    def __len__(self):
        self.recalc()
        return len(self.objects)


helper_list = HelperList()



from kts.modelling.mixins import MultiClassifierMixin

try:
    from xgboost import XGBClassifier as _XGBC

    class XGBClassifier(MultiClassifierMixin, _XGBC):
        """ """

        short_name = "xgb"
        tracked_params = [
            "base_score",
            "booster",
            "colsample_bylevel",
            "colsample_bytree",
            "gamma",
            "learning_rate",
            "max_delta_step",
            "max_depth",
            "min_child_weight",
            "missing",
            "n_estimators",
            "objective",
            "reg_alpha",
            "reg_lambda",
            "scale_pos_weight",
            "subsample",
        ]

except ImportError:
    pass

try:
    from lightgbm import LGBMClassifier as _LGBMC

    class LGBMClassifier(MultiClassifierMixin, _LGBMC):
        """ """

        short_name = "lgb"
        tracked_params = [
            "class_weight",
            "boost",
            "min_data_in_leaf",
            "bagging_freq",
            "subsample_freq",
            "bagging_fraction",
            "learning_rate",
            "reg_lambda",
            "feature_fraction",
            "min_sum_hessian_in_leaf",
            "boosting_type",
            "subsample_for_bin",
            "reg_alpha",
            "tree_learner",
            "colsample_bytree",
            "min_child_weight",
            "min_child_samples",
            "subsample",
            "max_depth",
            "boost_from_average",
            "num_leaves",
            "objective",
            "n_estimators",
            "min_split_gain",
        ]

except ImportError:
    pass

try:
    from catboost import CatBoostClassifier as _CBC

    class CatBoostClassifier(MultiClassifierMixin, _CBC):
        """ """

        short_name = "cb"
        tracked_params = [
            "iterations",
            "learning_rate",
            "l2_leaf_reg",
            "bootstrap_type",
            "bagging_temperature",
            "depth",
            "one_hot_max_size",
            "leaf_estimation_method",
            "nan_mode",
            "feature_border_type",
            "border_count",
            "max_depth",
            "colsample_bylevel",
            "eval_metric",
        ]

except ImportError:
    pass

from sklearn.ensemble import RandomForestClassifier as _RFC


class RandomForestClassifier(MultiClassifierMixin, _RFC):
    """ """

    short_name = "rf"
    tracked_params = [
        "bootstrap",
        "class_weight",
        "criterion",
        "max_depth",
        "max_features",
        "max_leaf_nodes",
        "min_impurity_decrease",
        "min_impurity_split",
        "min_samples_leaf",
        "min_samples_split",
        "min_weight_fraction_leaf",
        "n_estimators",
        "oob_score",
    ]


from sklearn.linear_model import LogisticRegression as _LR


class LogisticRegression(MultiClassifierMixin, _LR):
    """ """

    short_name = "lr"
    tracked_params = ["C", "penalty", "fit_intercept", "class_weight"]



import sys
from kts.cli.utils import get_mode

STORAGE_PATH = "../storage/"
INDEX_COLUMN_PREFIX = "__kts__index_"
IS_PREVIEW_CALL = 0
MEMORY_LIMIT = 4 * (1024 ** 3)  # 4 Gb
GOAL = "MAXIMIZE"
MAX_LEN_SOURCE = 100

# CACHE_MODE = 'disk_and_ram'  # "disk", "disk_and_ram", "ram"
# CACHE_POLICY = 'everything'  # "everything", "service"

CACHE_MODE, CACHE_POLICY, ROOT_DIR = get_mode()

LB_DF_NAME = "__leaderboard"

SERVICE_DF_NAMES = [LB_DF_NAME]


if CACHE_MODE in ["disk", "disk_and_ram"]:
    sys.path.insert(0, ROOT_DIR)
    try:
        import kts_config as user_config
        from kts_config import *
    except:
        pass

# So dumb. I know.
# TODO: implement all this ... with pathlib
if not STORAGE_PATH.endswith("/"):
    STORAGE_PATH += "/"
